{=====================================================================}
{ Base64 unit version 1.0.3 for Borland Delphi 2.0                    }
{                                                                     }
{ http://www.kobira.co.jp/sakura/windows.htm                          }
{ author Seiji-Yasuda seiji@kobira.co.jp                              }
{ Document:                                                           }
{  RFC1521,RFC1522,DWJ<u8106117@cseserv.cse.ttit.edu.tw> mimecode.pas }
{                                                                     }
{ You may use it freely at your own risk in any kind of environment.  }
{                                                                     }
{ Note:                                                               }
{  16/Apr./'97 version 1.0.0  local release.                          }
{  10/Apr./'97 version 1.0.1                                          }
{   normal unit --> class unit                                        }
{  20/Apr./'97 version 1.0.2                                          }
{   bug fix!                                                          }
{  9/Mar./'97 version 1.0.3                                           }
{   freeware release                                                  }
{                                                                     }
{=====================================================================}

unit base64;

interface

uses
  SysUtils;

type
  TBase64 = class
  private
    class function EncodeByte(c: Byte): Byte;
    class function DecodeByte(c: Byte): Byte;
    class procedure Encode3to4(d, s: PChar);
    class procedure Decode4to3(d, s: PChar);
  public
    class function Encode(s: PChar; sn: integer; d: PChar; var dn: integer; flg: boolean): boolean;
    class function Decode(s: PChar; sn: integer; d: PChar; var dn: integer; flg: boolean): boolean;
  end;

implementation

//=============================================================
// encode string for base64.
//  s   : (in) original data
//  sn  : (in) original data length
//  d   : (out)return encode data
//  dn  : (out)encode data length
//  flg : (in) true:end data
//=============================================================

class function TBase64.Encode(s: PChar; sn: integer; d: PChar;
  var dn: integer; flg: boolean): boolean;
var
  h, n: integer;
  tmp: array[0..7] of Char;
begin
  h := (sn mod 3);
  if (h <> 0) and (not flg) then
    begin
      Result := False;
      exit;
    end;
  n := 0;
  dn := 0;
  while n < (sn - h) do
    begin
      Encode3to4(@d[dn], @s[n]);
      n := n + 3;
      dn := dn + 4;
    end;
  if h <> 0 then
    begin
      case h of
        1:
          begin
            tmp[0] := s[n + 0];
            tmp[1] := Char(0);
            tmp[2] := Char(0);
            Encode3to4(@d[dn], @tmp[0]);
            d[dn + 2] := Char('=');
            d[dn + 3] := Char('=');
          end;
        2:
          begin
            tmp[0] := s[n + 0];
            tmp[1] := s[n + 1];
            tmp[2] := Char(0);
            Encode3to4(@d[dn], @tmp[0]);
            d[dn + 3] := Char('=');
          end;
      end;
      dn := dn + 4;
    end;
  Result := True;
end;

//=============================================================
// decode string for base64.
//  s   : (in) encode data
//  sn  : (in) encode data length
//  d   : (out)return original data
//  sn  : (out)original data length
//  flg : (in) true:end data
//=============================================================

class function TBase64.Decode(s: PChar; sn: integer; d: PChar;
  var dn: integer; flg: boolean): boolean;
var
  n, h: integer;
  tmp: array[0..7] of Char;
begin
  n := 0;
  dn := 0;
  while n < sn do
    begin
      h := 0;
      StrLCopy(@tmp, @s[n], 4);
      if s[n + 3] = '=' then
        begin
          Inc(h);
          if s[n + 2] = '=' then
            Inc(h);
        end;
      Decode4to3(@d[dn], @tmp);
      n := n + 4;
      dn := (dn + 3) - h;
    end;
  Result := True;
end;

//-------------------------------------------------------
// encode character for base64.
//

class function TBase64.EncodeByte(c: Byte): Byte;
begin
   // 'A'-'Z'
  if (Byte(0) <= c) and (c <= Byte(25)) then
    Result := Byte(Byte('A') + c)
   // 'a'-'z'
  else
    if (Byte(26) <= c) and (c <= Byte(51)) then
      Result := Byte(Byte('a') + c - 26)
   // '0'-'9'
    else
      if (Byte(52) <= c) and (c <= Byte(61)) then
        Result := Byte(Byte('0') + c - 52)
   // '+'
      else
        if (Byte(62) = c) then
          Result := Byte('+')
   // '/'
        else
          if (Byte(63) = c) then
            Result := Byte('/')
   // 64 over
          else
            Result := c;
end;

//-------------------------------------------------------
// decode character for base64.
//

class function TBase64.DecodeByte(c: Byte): Byte;
begin
   // 0-25
  if (Byte('A') <= c) and (c <= Byte('Z')) then
    Result := Byte(c - Byte('A'))
   // 26-51
  else
    if (Byte('a') <= c) and (c <= Byte('z')) then
      Result := Byte(26 + c - Byte('a'))
   // 52-61
    else
      if (Byte('0') <= c) and (c <= Byte('9')) then
        Result := Byte(52 + c - Byte('0'))
   // 62
      else
        if (Byte('+') = c) then
          Result := Byte(62)
   // 63
        else
          if (Byte('/') = c) then
            Result := Byte(63)
   // 64 over
          else
            Result := c;
end;

//-------------------------------------------------------
// encode 3 character to 4 character for base64.
//

class procedure TBase64.Encode3to4(d, s: PChar);
var
  b: Byte;
begin
   // XXXX-XX00 0000-0000 0000-0000
  b := Byte(ord(s[0]) shr 2);
  b := EncodeByte(b);
  d[0] := Char(b);
   // 0000-00XX XXXX-0000 0000-0000
  b := Byte(((ord(s[0]) and $3) shl 4) + (Ord(s[1])) shr 4);
  b := EncodeByte(b);
  d[1] := Char(b);
   // 0000-0000 0000-XXXX XX00-0000
  b := Byte(((Ord(s[1]) and $F) shl 2) + (Ord(s[2]) and $C0) shr 6);
  b := EncodeByte(b);
  d[2] := Char(b);
   // 0000-0000 0000-0000 00XX-XXXX
  b := Byte((Ord(s[2]) and $3F));
  b := EncodeByte(b);
  d[3] := Char(b);
end;

//-------------------------------------------------------
// decode 4 character to 3 character for base64.
//

class procedure TBase64.Decode4to3(d, s: PChar);
var
  b1, b2, b3, b4: Byte;
begin
   // XXXX-XX00 0000-0000 0000-0000
  b1 := DecodeByte(Byte(s[0]));
   // 0000-00XX XXXX-0000 0000-0000
  b2 := DecodeByte(Byte(s[1]));
   // 0000-0000 0000-XXXX XX00-0000
  b3 := DecodeByte(Byte(s[2]));
   // 0000-0000 0000-0000 00XX-XXXX
  b4 := DecodeByte(Byte(s[3]));
  d[0] := Char((Ord(b1) shl 2) + (Ord(b2) shr 4));
  d[1] := Char((Ord(b2) shl 4) + (Ord(b3) shr 2));
  d[2] := Char((Ord(b3) shl 6) + (Ord(b4) and $3F));
end;

end.

